# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

"""The interactive NVIDIA-GPU process viewer."""

import sys

from nvitop.cli import main


if __name__ == '__main__':
    sys.exit(main())
